////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2011-2012 Kentaro Sekimoto  All rights reserved.
////////////////////////////////////////////////////////////////////////////

#include <tinyhal.h>
#include "..\FM3.h"
//--//

// ADCR : A/D Control Register
// ADSR : A/D Status Register
// SCCR : Scan Conversion Control Register
// SFNS : Scan Conversion FIFO Configuration Register
// SCFD : Scan Conversion FIFO Data Register
// SCIS : Scan Conversion Input Selection Register
// PCCR : Priority Conversion Control Register
// PFNS : Priority FIFO Number Selection Register
// PCFD : Priority Conversion FIFO Data Register
// PCIS : Priority Conversion Input Selection Register
// CMPD : A/D Comparison configuration register
// CMPCR: A/D Comparison Control Register
// ADSS : A/D Sampling time Selection register
// ADST : A/D Sampling Time configuration register
// ADCT : A/D Compare Time configuration register
// ADCEN: A/D Conversion Enable register

/*
 *  Register Bit Definition
 */
#define ADCR_OVRIE           0x01U
#define ADCR_CMPIE           0x02U
#define ADCR_PCIE            0x04U
#define ADCR_SCIE            0x08U
#define ADCR_CMPIF           0x20U
#define ADCR_PCIF            0x40U
#define ADCR_SCIF            0x80U
#define ADCR_IFALL           (ADCR_CMPIF | ADCR_PCIF | ADCR_SCIF)

#define ADSR_SCS             0x01U
#define ADSR_PCS             0x02U
#define ADSR_PCNS            0x04U
#define ADSR_FDAS            0x40U
#define ADSR_ADSTP           0x80U

#define SCCR_SSTR            0x01U
#define SCCR_SHEN            0x02U
#define SCCR_RPT             0x04U
#define SCCR_SFCLR           0x10U
#define SCCR_SOVR            0x20U
#define SCCR_SFUL            0x40U
#define SCCR_SEMP            0x80U
#define SCFDL_INVL           0x1000U

#define PCCR_PSTR            0x01U
#define PCCR_PHEN            0x02U
#define PCCR_PEEN            0x04U
#define PCCR_ESCE            0x08U
#define PCCR_PFCLR           0x10U
#define PCCR_POVR            0x20U
#define PCCR_PFUL            0x40U
#define PCCR_PEMP            0x80U
#define PCFDL_INVL           0x1000U

#define CMPCR_CMD0           0x20U
#define CMPCR_CMD1           0x40U
#define CMPCR_CMPEN          0x80U

#define ADCEN_ENBL           0x01U
#define ADCEN_READY          0x02U

struct FM3_AD_Driver
{
#ifdef MB9BF618T
    static const UINT32 c_Channels = 32;
#else
    static const UINT32 c_Channels = 16;
#endif
    static const UINT8 c_Pins[c_Channels];
};

#ifdef MB9BF618T
const UINT8 FM3_AD_Driver::c_Pins[FM3_AD_Driver::c_Channels] =
{
        FM3__GPIO::P10, FM3__GPIO::P11, FM3__GPIO::P12, FM3__GPIO::P13,
        FM3__GPIO::P14, FM3__GPIO::P15, FM3__GPIO::P16, FM3__GPIO::P17,
        FM3__GPIO::P18, FM3__GPIO::P19, FM3__GPIO::P1A, FM3__GPIO::P1B,
        FM3__GPIO::P1C, FM3__GPIO::P1D, FM3__GPIO::P1E, FM3__GPIO::P1F,
        FM3__GPIO::PB0, FM3__GPIO::PB1, FM3__GPIO::PB2, FM3__GPIO::PB3,
        FM3__GPIO::PB4, FM3__GPIO::PB5, FM3__GPIO::PB6, FM3__GPIO::PB7,
        FM3__GPIO::PB8, FM3__GPIO::PB9, FM3__GPIO::PBA, FM3__GPIO::PBB,
        FM3__GPIO::PBC, FM3__GPIO::PBD, FM3__GPIO::PBE, FM3__GPIO::PBF
};
#else
const UINT8 FM3_AD_Driver::c_Pins[FM3_AD_Driver::c_Channels] =
{
        FM3__GPIO::P10, FM3__GPIO::P11, FM3__GPIO::P12, FM3__GPIO::P13,
        FM3__GPIO::P14, FM3__GPIO::P15, FM3__GPIO::P16, FM3__GPIO::P17,
        FM3__GPIO::P18, FM3__GPIO::P19, FM3__GPIO::P1A, FM3__GPIO::P1B,
        FM3__GPIO::P1C, FM3__GPIO::P1D, FM3__GPIO::P1E, FM3__GPIO::P1F
};
#endif

FM3_AD_Driver g_FM3_AD_Driver;

BOOL AD_Initialize(ANALOG_CHANNEL channel, INT32 precisionInBits)
{
    UINT32 ch = (UINT32)channel;
    if (ch >= g_FM3_AD_Driver.c_Channels)
        return FALSE;
    if (precisionInBits == 12) {
        if (ch < 16)
            FM3_GPIO->PFR1 |= (1 << ch);
        else
            FM3_GPIO->PFRB |= (1 << (ch-16));
        FM3_GPIO->ADE |= (1 << ch);
        FM3_ADC0->ADCT = 3;
        FM3_ADC0->ADCEN = ADCEN_ENBL;
        while((FM3_ADC0->ADCEN & ADCEN_READY) != ADCEN_READY) ;
        if (ch < 8) {
            FM3_ADC0->SCIS0 |= (1 << ch);
        } else if (ch < 16) {
            FM3_ADC0->SCIS1 |= (1 << (ch-8));
        } else if (ch < 24) {
            FM3_ADC0->SCIS2 |= (1 << (ch-16));
        } else {
            FM3_ADC0->SCIS3 |= (1 << (ch-24));
        }
        FM3_ADC0->ADST0 = 0xA9;     // STX:6=x64, ST:9=x(9+1),  -> 64 x (9 + 1) + 1 = 641
        FM3_ADC0->ADST1 = 0xA9;     // PCLK 72MHz=13.9ns -> 13.9ns x 641 = 8910ns
        FM3_ADC0->CMPCR = ch;
        FM3_ADC0->ADSR = ADSR_FDAS; // ADSTP(R/W):0, FDAS(R/W):1(LSB), , , , PCNS(R):0, PCS(R):0, SCS(R):0
        FM3_ADC0->ADCR = 0;         // SCIF:0, PCIF:0, CMPIF:0, , SCIE:0 , PCIE:0 , CMPIE:0, OVRIE:0
        FM3_ADC0->SCCR |= SCCR_SFCLR;
        FM3_ADC0->SCCR &= ~SCCR_RPT;
        return TRUE;
    } else
        return FALSE;
}

INT32 AD_Read(ANALOG_CHANNEL channel)
{
    volatile UINT32 value32;
    FM3_ADC0->SCCR |= SCCR_SSTR;    // A/D start
    do {
        value32 = FM3_ADC0->SCFD;
    } while (((value32 & SCFDL_INVL) != 0) || ((UINT32)channel != (value32 & 0x1F)));
    FM3_ADC0->SCCR = (FM3_ADC0->SCCR | SCCR_SFCLR) & ~(SCCR_RPT | SCCR_SOVR);    // A/D stop
    return (value32 >> 16);
}

UINT32 AD_ADChannels()
{
    return FM3_AD_Driver::c_Channels;
}

GPIO_PIN AD_GetPinForChannel(ANALOG_CHANNEL channel)
{
    if ((UINT32)channel >= g_FM3_AD_Driver.c_Channels)
        return GPIO_PIN_NONE;
    return (GPIO_PIN)g_FM3_AD_Driver.c_Pins[(UINT32)channel];
}

BOOL AD_GetAvailablePrecisionsForChannel(ANALOG_CHANNEL channel, INT32* precisions, UINT32& size)
{
    size = 0;
    if (precisions == NULL || (UINT32)channel >= g_FM3_AD_Driver.c_Channels)
        return FALSE;
    precisions[0] = 12;
    size = 1;
    return TRUE;
}
